/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <stdlib.h>
#include "minrtsFftUtil.h"

minrtsFft1d *minrtsMakeFft1d(float gD[3], float cP[2]) {
/* ******************************************************************
 * make an fft object from explicit supplied data
 * Inputs:
 *         gD  & parameters describing array onto which data
 *             & is gridded: initial value, increment, number of points
 *         cP  & parameters describing gaussian used for convolving on
 *             & grid: sigma, maximum extent (grid pixels)
 *             & if 0, defaults are taken
 * Returns:    & Fft object
 ******************************************************************* */
  minrtsFft1d* ff;
  int i, nX;
  float twopi, x;
  ff= (minrtsFft1d *)malloc(sizeof(minrtsFft1d));
  for (i=0; i<3; i++) ff->gridData[i] = gD[i];
  ff->nX = gD[2];
  nX = gD[2];
  ff->grid     = (float *)malloc(nX * sizeof(float)); /* gridded data */
  ff->grid2    = (float *)malloc(nX * sizeof(float)); /* 2nd gridded data */
  ff->kGrid    = (float *)malloc(nX * sizeof(float)); /* coordinates in fft space */
  ff->gridCorr = (float *)malloc(nX * sizeof(float)); /* gridding correction function */
  ff->filter   = (float *)malloc(2 * nX * sizeof(float)); /* complex filter */
  ff->work     = (float *)malloc(2 * nX * sizeof(float));
  ff->cWork    = (float *)malloc(2 * nX * sizeof(float));
/* set kGrid values based on gridding parameters */
  twopi = 2.* M_PI;
  for (i=0; i<nX; i++) ff->kGrid[i] =
                         (i - .5*gD[2]) * twopi/(gD[1]*gD[2]);
/* set convolving parameters including possible defaults */
  if (cP == 0) {
    ff->convParm[0] = 0.85;
    ff->convParm[1] = 3;
  } else {
    ff->convParm[0] = cP[0];
    ff->convParm[1] = cP[1];
  }
/*  correct user if he is nuts */
  if (ff->convParm[0] < 0.5) ff->convParm[0] = 0.5;
  if (ff->convParm[1] < 2) ff->convParm[0] = 2;
/* set convolution correction function */
  for (i=0; i< nX; i++) {
    x = ff->convParm[0]* ((float)i-.5*nX) * twopi/nX;
    x = .5*x*x;
    if (x > 5) x=5;
    if (i >= nX/2) ff->gridCorr[i-nX/2] = exp(x) ; else
      ff->gridCorr[nX/2+i] = exp(x);
  };
/* set default filter */
  for (i=0; i<nX; i++)  ff->filter[i] = 1.;
  return ff;
} /* end of minrtsMakeFft1d */

void minrtsKillFft1d (minrtsFft1d *ff) {
/* ******************************************************************
 * free memory associated with fft object
 * Inputs:
 *         ff  & fft object
 ********************************************************************* */
  free (ff->grid);
  free (ff->grid2);
  free (ff->kGrid);
  free (ff->gridCorr);
  free (ff->work);
  free (ff->cWork);
  free (ff->filter);
} /* end of minrtsKillFft1d */

float *minrtsFftGrid(minrtsFft1d *ff) {
/* ******************************************************************
 * return data already gridded in object
 * Inputs:
 *         ff  & fft object
 * Returns:
 *             & gridded copy of data
 ******************************************************************* */
  int i;
  float *out;
  out = (float *)malloc(ff->nX * sizeof(float));
  for (i=0; i<ff->nX; i++) out[i] = ff->grid[i];
  return out;
}

float *minrtsFftKGrid(minrtsFft1d *ff) {
/* ******************************************************************
 * return coordinate values of k-array (fft of input grid array)
 * Inputs:
 *         ff  & fft object
 * Returns:
 *             & k-values
 ******************************************************************* */
  int i;
  float *out;
  out = (float *)malloc(ff->nX * sizeof(float));
  for (i=0; i<ff->nX; i++) out[i] = ff->kGrid[i];
  return out;
}

float *minrtsFftCoord(minrtsFft1d *ff) {
/* ******************************************************************
 * return coordinates of input grid points (generated by cP values)
 * Inputs:
 *         ff  & fft object
 * Returns:
 *             & coordinate values
 ******************************************************************* */
  int i;
  float *out;
  out = (float *)malloc(ff->nX * sizeof(float));
  for (i=0; i<ff->nX; i++) out[i] = ff->gridData[0] +
                             ff->gridData[1] * i;
  return out;
}
float *minrtsFftGridCorr(minrtsFft1d *ff) {
/* ******************************************************************
 * return function that corrects for gridding after fft
 * Inputs:
 *         ff  & fft object
 * Returns:
 *             & correction function
 ******************************************************************* */
  int i;
  float *out;
  out = (float *)malloc(ff->nX * sizeof(float));
  for (i=0; i<ff->nX; i++) out[i] = ff->gridCorr[i];
  return out;
}

void minrtsFftSetRFilter(minrtsFft1d *ff, float *filter) {
/* ******************************************************************
 * insert REAL k-space filter in fft object
 * abscissa values are given by ff->kgrid
 * number of data values must agree
 * Inputs:
 *         ff  & fft object
 *      filter & values to insert
 ******************************************************************* */
  int i;
  for (i=0; i< (ff->nX); i++) {
     ff->filter[2*i] = filter[i];/* real -> complex */
     ff->filter[2*i+1] = 0.;
  }
  return;
}

void minrtsFftSetCFilter(minrtsFft1d *ff, minrtsFilter *filter) {
/* ******************************************************************
 * insert COMPLEX k-space filter in fft object from filter object
 * abscissa values are given by ff->kgrid
 * Inputs:
 *         ff  & fft object
 *      filter & object containing to insert
 ******************************************************************* */
  int i;
  float *cTrans;
  cTrans = minrtsFilterComplexFilter(filter, ff->nX, ff->kGrid);
  for (i=0; i< 2*(ff->nX); i++) ff->filter[i] = cTrans[i];
  free(cTrans);
  return;
}

void minrtsFftMultCFilter(minrtsFft1d *ff, minrtsFilter *filter) {
/* ******************************************************************
 * Multiply COMPLEX k-space filter in fft object from filter object
 * time current filter.
 * abscissa values are given by ff->kgrid
 * Inputs:
 *         ff  & fft object
 *      filter & object containing to insert
 ******************************************************************* */
   int i;
   float x;
   float *cTrans;
   cTrans = minrtsFilterComplexFilter(filter, ff->nX, ff->kGrid);
   for (i=0; i< (ff->nX); i++) {
      x = ff->filter[2*i]*cTrans[2*i] - ff->filter[2*i+1]*cTrans[2*i+1];
      ff->filter[2*i+1] = ff->filter[2*i]*cTrans[2*i+1] + ff->filter[2*i+1] * cTrans[2*i];
      ff->filter[2*i] = x;
   }
   free(cTrans);
   return;
}


void minrtsFftGridData (minrtsFft1d *ff, int nData, float *xPos, 
                        float *data, float *data2, float *dataWeight) {
/* ******************************************************************
   * take one or two arrays of real data values at given positions and
   * grid them by convolution with a gaussian onto internal grids: grid an grid2
   * An array of weights may also be provided, if not a NULL (0) should
   * be put in this array in which case the weights are implicitly all = 1
   * An attempt is made to provide "uniform weighting": at each grid point the
   * sum of the weighted data points is divided by the sum of the weights
   * Inputs:
   *         ff    & fft object
   *         nData & number of input data points: must be same for both arrays & weights
   *         xPos  & positions of input points
   *         data  & input data values
   *         data2 & 2nd array, same x-values, put a NULL here if no 2nd array
   *       dataweights & weights for data, NULL if none desired
    ******************************************************************* */
  int i, nX, dW, doPoint, xC;
  float sigma, dx, x;
  float xCoord, *xWt, xTot = 0, *gridWt;
  int gMax, xMin, xMax, iG;
  nX = ff->nX;
/* set previous data to zero */
  for (i=0; i< nX; i++) ff->grid[i] = 0.;
  sigma  = 0.5/(ff->convParm[0] * ff->convParm[0]); /* copy stored parameters into local vars */
  dx     = ff->convParm[1];
  gMax   = nX - 1;
/* get work arrays to store weighting values */
  xWt    = ff->work;
  gridWt = ff->cWork;
  for (i=0; i< nX; i++) *(gridWt+i) = 0.;
/* dataWeight is not set */
  dW = (dataWeight == 0);
/* calculate weights to approximate uniform weights */
  if (!dW) {
     for (i=0; i<nData; i++) { 
        if (dataWeight[i] > 0.) {
           xC=  floor(.5+(xPos[i] - ff->gridData[0])/ff->gridData[1]);
           if ((xC >= 0) && (xC <(ff->gridData[2])))
              gridWt[xC] = gridWt[xC] + dataWeight[i];
        }
     }
  } else {
     for (i=0; i<nData; i++) { 
        xC= floor(.5+(xPos[i] - ff->gridData[0])/ff->gridData[1]);
        if ((xC >= 0) && (xC <(ff->gridData[2]))) gridWt[xC]++;
     }
  }
  for (i=0; i<nData; i++) { /* loop over input points */
/* skip points if zero weight specified */
    doPoint = dW;     /* if not dataWeight, do every point */
    if (!dW) doPoint = (dataWeight[i] != 0);
    xCoord = (xPos[i] - ff->gridData[0])/ff->gridData[1];
    xC = floor(.5+xCoord);
    doPoint = doPoint && (xC >= 0) && (xC <(ff->gridData[2]));
    if (doPoint) {
/* calculate normalized coordinates of input point */
/* calculate range of output points on grid */
      xMin   = floor(xCoord - dx);
      if (xMin < 0) xMin = 0;
      xMax = ceil(xCoord + dx);
      if(xMax > gMax) xMax = gMax;
/* only worry about points inside of grid */
      if ((xCoord >= 0.) && (xCoord <= gMax ))
      {
/* calculate weight array for output points */
         xTot = 0.;
         for (iG=xMin; iG<=xMax; iG++)
         {
           x = (iG - xCoord);
           xWt[iG] = exp(-sigma * x * x);
           xTot = xTot + xWt[iG];
         }
/* map input point onto output points, incrementing weights */
         for (iG=xMin; iG<=xMax; iG++) 
            ff->grid[iG] = ff->grid[iG]+ data[i] * xWt[iG]/(xTot*gridWt[xC]);
         if (data2 != NULL)
             for (iG=xMin; iG<=xMax; iG++) 
               ff->grid2[iG] = ff->grid[iG]+ data2[i] * xWt[iG]/(xTot*gridWt[xC]);
      }  /*  point inside gridding array */
    }     /*  if doPoint */
  }     /*   loop over input points */
  return;
} /* end of GridData */


void minrtsFftGridDataC (minrtsFft1d *ff, int nData, float *xPos, 
                        float *data, float *dataWeight) {
/* ******************************************************************
   * take one array of complex data values at given positions and
   * grid them by convolution with a gaussian onto internal grids: grid an grid2
   * An array of weights may also be provided, if not a NULL (0) should
   * be put in this array in which case the weights are implicitly all = 1
   * An attempt is made to provide "uniform weighting": at each grid point the
   * sum of the weighted data points is divided by the sum of the weights
   * Inputs:
   *         ff    & fft object
   *         nData & number of input data points: must be same for both arrays & weights
   *         xPos  & positions of input points
   *         data  & input data values
   *       dataweights & weights for data, NULL if none desired
    ******************************************************************* */
  int i, nX, dW, doPoint, xC;
  float sigma, dx, x;
  float xCoord, *xWt, xTot = 0, *gridWt;
  int gMax, xMin, xMax, iG;
  nX = ff->nX;
/* set previous data to zero */
  for (i=0; i< nX; i++) {
     ff->grid[i] = 0.;
     ff->grid2[i] = 0.;
     ff->work[i] = 0.;
     ff->cWork[i] = 0.;
  }
  sigma  = 0.5/(ff->convParm[0] * ff->convParm[0]); /* copy stored parameters into local vars */
  dx     = ff->convParm[1];
  gMax   = nX - 1;
/* get work arrays to store weighting values */
  xWt    = ff->work;
  gridWt = ff->cWork;
/* dataWeight is not set */
  dW = (dataWeight == 0);
/* calculate weights to approximate uniform weights */
  if (!dW) {
     for (i=0; i<nData; i++) { 
        if (dataWeight[i] > 0.) {
           xC=  floor(.5+(xPos[i] - ff->gridData[0])/ff->gridData[1]);
           if ((xC >= 0) && (xC <(ff->gridData[2])))
              gridWt[xC] = gridWt[xC] + dataWeight[i];
        }
     }
  } else {
     for (i=0; i<nData; i++) { 
        xC= floor(.5+(xPos[i] - ff->gridData[0])/ff->gridData[1]);
        if ((xC >= 0) && (xC <(ff->gridData[2]))) gridWt[xC]++;
     }
  }
  for (i=0; i<nData; i++) { /* loop over input points */
/* skip points if zero weight specified */
    doPoint = dW;     /* if not dataWeight, do every point */
    if (!dW) doPoint = (dataWeight[i] != 0);
    xCoord = (xPos[i] - ff->gridData[0])/ff->gridData[1];
    xC = floor(.5+xCoord);
    doPoint = doPoint && (xC >= 0) && (xC <(ff->gridData[2]));
    if (doPoint) {
/* calculate normalized coordinates of input point */
/* calculate range of output points on grid */
      xMin   = floor(xCoord - dx);
      if (xMin < 0) xMin = 0;
      xMax = ceil(xCoord + dx);
      if(xMax > gMax) xMax = gMax;
/* only worry about points inside of grid */
      if ((xCoord >= 0.) && (xCoord <= gMax ))
      {
/* calculate weight array for output points */
        xTot = 0.;
        for (iG=xMin; iG<=xMax; iG++)
        {
          x = (iG - xCoord);
          xWt[iG] = exp(-sigma * x * x);
          xTot = xTot + xWt[iG];
        }
/* map input point onto output points, incrementing weights */
        if (dW) 
        {
          for (iG=xMin; iG<=xMax; iG++) 
          { 
            x = xWt[iG]/(xTot*gridWt[xC]);
            ff->grid[iG]  = ff->grid[iG]  + data[2*i]   * x;
            ff->grid2[iG] = ff->grid2[iG] + data[2*i+1] * x;
          }
        } 
        else 
        {
          for (iG=xMin; iG<=xMax; iG++)  
          {
            x = xWt[iG]/(xTot*gridWt[xC]);
            ff->grid[iG]  = ff->grid[iG]  + dataWeight[i]*data[2*i]  * x;
            ff->grid2[iG] = ff->grid2[iG] + dataWeight[i]*data[2*i+1] * x;
          }
        } /* loop over output points */
      }  /*  point inside gridding array */
    }     /*  if doPoint */
  }     /*   loop over input points */

  return;
} /* end of GridDataC */

void minrtsFftGridFft (minrtsFft1d *ff, int nData, float *xPos, float *data,
   float *dataWeight, float *fft){
/* ******************************************************************
   * take one array of real data values at given positions and
   * grid them by convolution with a gaussian onto internal grids: then
   * do an fft on the gridded data and correct for the gridding effect.
   * Inputs:
   *         ff    & fft object
   *         nData & number of input data points: must be same for both arrays & weights
   *         xPos  & positions of input points
   *         data  & input data values
   *       dataweights & weights for data, NULL if none desired
   * Outputs :
   *         fft   & output complex array, assumed already allocated
   ******************************************************************** */
   int i, j, nX;
   float twopi = 6.2831853072;
   float gC;
   float zpR, zpI, zp;
   minrtsFftGridData(ff, nData, xPos, data, 0, dataWeight); 
   nX = ff->nX;
/* zero output */
   for (i=0; i<2*nX; i++) fft[i] = 0.;
/* convert real input data to complex */
   for (i=0; i<nX; i++) {
      ff->cWork[2*i]   = ff->grid[i];
      ff->cWork[2*i+1] = 0.;
   }
/* only one input */
   fourg (ff->cWork, nX, 1, ff->work); /* fft */
/* correct for gridding, shift to correct zero points and correct phase */
   zp = twopi * ff->gridData[0]/(nX * ff->gridData[1]); 
   for (i=0; i<nX; i++) {
      j = i + nX/2;
      if (j >= nX) j=j - nX;
      zpR = cos(zp * (i -nX/2));
      zpI = sin(zp * (i -nX/2));
      gC = ff->gridCorr[i];
/*      if (gC > 15) gC = 15; */
        if (gC > 5) gC = 5; 
      fft[2*j]   = gC * (ff->cWork[2*i]*zpR - ff->cWork[2*i+1]*zpI);
      fft[2*j+1] = gC * (ff->cWork[2*i]*zpI + ff->cWork[2*i+1]*zpR);
   }
   return ;
} /* end of GridFft */


void minrtsFftGridFftC (minrtsFft1d *ff, int nData, float *xPos, float *data,
   float *dataWeight, float *fft){
/* ******************************************************************
   * take one array of complex data values at given positions and
   * grid them by convolution with a gaussian onto internal grids: then
   * do an fft on the gridded data and correct for the gridding effect.
   * Inputs:
   *         ff    & fft object
   *         nData & number of input data points: must be same for both arrays & weights
   *         xPos  & positions of input points
   *         data  & input data values
   *       dataweights & weights for data, NULL if none desired
   *       fft     & output array, previously allocated
   ******************************************************************** */
   int i, j, nX;
   float twopi = 6.2831853072;
   float gC;
   float zpR, zpI, zp;
   minrtsFftGridDataC(ff, nData, xPos, data, dataWeight); 
   nX = ff->nX;
/* zero output */
   for (i=0; i<2*nX; i++) fft[i] = 0.;
/* 
      convert two real input arrays to complex 
*/
   for (i=0; i<nX; i++) {
      ff->cWork[2*i]   = ff->grid[i];
      ff->cWork[2*i+1] = ff->grid2[i];
   }
   fourg (ff->cWork, nX, 1, ff->work); /* fft */
/* 
      correct for gridding, shift to correct zero points and correct phase 
*/
   zp = twopi * ff->gridData[0]/(nX * ff->gridData[1]);
   for (i=0; i<nX; i++) {
      j = i + nX/2;
      if (j >= nX) j=j - nX;
      zpR = cos(zp * (i -nX/2));
      zpI = sin(zp * (i -nX/2));
      gC = ff->gridCorr[i];
      if (gC > 15) gC = 15;
      fft[2*j] = gC * (ff->cWork[2*i]*zpR - ff->cWork[2*i+1]*zpI);
      fft[2*j+1] = gC * (ff->cWork[2*i]*zpI + ff->cWork[2*i+1]*zpR);
   }
} /* end of GridFft Complex */

void minrtsFftBackFft (minrtsFft1d *ff, float *inData, int oType, float *outData) {
/* *****************************************************************************
   * reverse Fft using same parameters as stored in object 
   * Inputs:  inData  &  data in k-space (Complex)
   *          oType   &  output Type:
   *                  &  0 = COMPLEX
   *                  &  1 = REAL
   *                  &  2 = IMAGINARY
   *                  &  3 = AMP
   *                  &  4 = AMP*2
   * Outputs  outData &  data in x-space
   ****************************************************************************** */
   int i, j, nX;
   j = 0;
   nX = ff->nX;
/*  
   shift left for fourg 
*/
   for (i=0; i<2*nX; i++) {
      j = i - nX;
      if (j<0) j=j+2*nX;
      ff->cWork[j] = inData[i];
   }
   fourg (ff->cWork, nX, -1, ff->work); /* fft */
/*  
   shift right for fourg and move to output
*/
   if (oType == 0) {
      for (i=0; i<2*nX; i++) {
         j = i - nX;
         if (j<0) j=j+2*nX;
         outData[j] = ff->cWork[i];
      }
   } else if (oType == 1) {
      for (i=0; i<nX; i++) {
         j = i - nX/2;
         if (j<0) j=j+nX;
         outData[j] = ff->cWork[2*i];
      }
   } else if (oType == 2) {
      for (i=0; i<nX; i++) {
         j = i - nX/2;
         if (j<0) j=j+nX;
         outData[j] = ff->cWork[2*i+1];
      }
   } else if (oType == 3) {
      for (i=0; i<nX; i++) {
         j = i - nX/2;
         if (j<0) j=j+nX;
         outData[j] = sqrt(ff->cWork[2*i]*ff->cWork[2*i] +
            ff->cWork[2*i+1]*ff->cWork[2*i+1]);
      }
   } else {
      for (i=0; i<nX; i++) {
         if (j<0) j=j+nX;
         outData[j] = ff->cWork[2*i]*ff->cWork[2*i] +
            ff->cWork[2*i+1]*ff->cWork[2*i+1];
      }
   }
   return;
}

void minrtsFftApplyFilter(minrtsFft1d *ff, int fType, float *data) {
/* *****************************************************************************
   * apply internally stored complex filter to data
   * Inputs:
   *       ff    & fft object
   *       fType & filter type:
   *             & 0 = complex, all values
   *             & 1 = complex only positive values of spatial frequency
   *             & 2 = abs values of filter
   * In/Outputs: & data
   *************************************************************************** */
   int i;
   float fR, fI, x;
   for (i=0; i<(ff->nX); i++) {
      if (fType == 0 || (ff->kGrid[i] > 0)) { /* normal filter */
         fR = ff->filter[2*i]; 
         fI = ff->filter[2*i+1];
      } else if (fType == 2) { /* absolute value of filter */
         fR = ff->filter[2*i];
         fI = ff->filter[2*i+1];
         fR = sqrt(fR*fR+fI*fI);
         fI = fR;
      } else { /* zero for negative spFreq */
         fR = 0.;
         fI = 0.;
      }
      x = data[2*i] * fR - data[2*i+1] * fI;
      data[2*i+1] = data[2*i+1] * fR  + data[2*i] * fI;
      data[2*i]   = x;
   }
   return;
}
void minrtsFftFilterData (minrtsFft1d *ff, int nData, float *xPos,
   float *data, float *dataWeight, int fType, int oType, float *outData) {
/* *****************************************************************************
   *   filter data with preinstalled filter by doing grid fft to
   *   k domain, multiplying by filter, and ffting back to x-domain
   * Inputs:
   *         ff   &  fft object
   *         nX   &  number of input data points
   *      xPos    &  xpositions of input data
   *      data    &  input data values
   *      fType   & filter Type
   *              & 0 all values
   *              & 1 positive frequencies only, others set to zero
   *              & 2 use absolute value of filter
   *      oType   & output Type
   *              & 0 = COMPLEX
   *              & 1 = REAL
   *              & 2 = IMAGINARY
   *              & 3 = ABS
   *              & 4 = SQUARE ABS
   *     output   & result, assumed already allocated
   *************************************************************************** */
   int i, j, nX;
   float *fftData;
   float tR, tI, zp, twopi, x;
   nX = ff->nX;
   twopi = 2.* M_PI;
   fftData = (float *)malloc(nX * 2 * sizeof(float));
   minrtsFftGridFft (ff, nData, xPos, data, dataWeight, fftData);
/* 
      apply filter and shift phase 
*/
   minrtsFftApplyFilter(ff, fType, fftData) ;
   zp = twopi * ff->gridData[0]/(nX * ff->gridData[1]);
   for (i=0; i<nX; i++) {
      tR =  cos(zp*(i-nX/2));
      tI = -sin(zp*(i-nX/2)); 
      x   = fftData[2*i] * tR - fftData[2*i+1]*tI;
      fftData[2*i+1] = fftData[2*i] * tI + fftData[2*i+1] * tR;
      fftData[2*i]   = x;
   }
/* 
      shift 1/2 cycle for fourg which expects 0freq at left edge
*/
   for (i=0; i<nX; i++) {
      x = fftData[i];
      j = i + nX;      
      fftData[i] = fftData[j];
      fftData[j] = x;
   }
   fourg (fftData, nX, -1, ff->work); /* fft */
   if (oType == 0) {
      for (i=0;i<2*nX;i++) outData[i] = fftData[i]/nX;
   } else if (oType == 1) {
      for (i=0;i<nX;i++) outData[i] = fftData[2*i]/nX;
   } else if (oType == 2) {
      for (i=0;i<nX;i++) outData[i] = fftData[2*i+1]/nX;
   } else if (oType == 3) {
      for (i=0;i<nX;i++) outData[i] = sqrt(
         fftData[2*i]*fftData[2*i] + fftData[2*i+1]*fftData[2*i+1])/nX;
   } else if (oType == 4) {
      for (i=0;i<nX;i++) outData[i] = 
         (fftData[2*i]*fftData[2*i] + fftData[2*i+1]*fftData[2*i+1])/nX;
   }
   free(fftData);
   return ;
} /* end of Filter */

void minrtsFftGridFft2 (minrtsFft1d *ff, int nData, float *xPos, 
   float *data, float *data2, float *dataWeight, float **fft1, float **fft2){
   int i, j, nX;
   float twopi ;
   float gC;
   float zpR, zpI, zp, *out1, *out2;
   minrtsFftGridData(ff, nData, xPos, data, data2, dataWeight); 
   twopi = 2.* M_PI;
   nX = ff->nX;
/* fft both gridded input vectors */
   minrtsDoubleRealFftF (nX, ff->grid, ff->grid2, &out1, &out2);
/* correct for gridding, shift to correct zero points and correct phase */
   zp = twopi * ff->gridData[0]/(nX * ff->gridData[1]);
   for (i=0; i<nX; i++) {
      j = i + nX/2;
      if (j >= nX) j=j - nX;
      zpR = cos(zp * (i -nX/2));
      zpI = sin(zp * (i -nX/2));
      gC = ff->gridCorr[j];
      if (gC > 15) gC = 15;
      out1[2*i]   = gC * (out1[2*i]*zpR - out1[2*i+1]*zpI);
      out1[2*i+1] = gC * (out1[2*i]*zpI + out1[2*i+1]*zpR);
      out2[2*i]   = gC * (out2[2*i]*zpR - out2[2*i+1]*zpI);
      out2[2*i+1] = gC * (out2[2*i]*zpI + out2[2*i+1]*zpR);
   }
   *fft1 = out1;
   *fft2 = out2;
   return ;
} /* end of GridFft2 */

float * minrtsMakeHermit (float *data, int nData) {
/* make data hmeritian by adding it to flipped conjugate */
/* assume zero point is in the middle */
   int i;
   float *output, tR, tI;
   output = (float *)malloc(2*nData*sizeof(float));
   for (i=1; i<nData; i++){
      tR = .5*(data[2*i] + data[2*(nData-i)]);
      tI = .5*(data[2*i+1] - data[2*(nData-i)+1]);
      output[2*i] = tR;
      output[2*i+1] = tI;
   }
   output[0] = data[0];
   output[1] = 0.;
   return output;
}

void minrtsDoubleRealFftF (int nData, float *real1, float *real2, 
   float **fft1, float **fft2 ) {
/* do two real forward ffts of the same length at the same time by putting one
   of them into the imaginary part of a complex number and use their
   hermitian properties to fish them out of the result */
/* note that the forward output has zero point in the middle, NOT AT EDGE */
   int i, j;
   float *cWork, *work, tR, tI;
   work  = (float *) malloc(2 * nData * sizeof(float));
   cWork = (float *) malloc(2 * nData * sizeof(float));
   for (i=0; i<nData; i++) {
      cWork[2*i]   = real1[i];
      cWork[2*i+1] = real2[i];
   }
   fourg (cWork, nData, 1, work); /* fft */
/* 
      shift zero point 
*/
   for (i=0; i<nData/2; i++) {
      j = i + nData/2;
      tR = cWork[2*i];
      tI = cWork[2*i+1];
      cWork[2*i]   = cWork[2*j];
      cWork[2*i+1] = cWork[2*j+1];
      cWork[2*j]   = tR;
      cWork[2*j+1] = tI;
   }
/*    
      fish out two parts by insisting on hermitian property 
*/
   *fft1 = minrtsMakeHermit (cWork, nData);
   for (i=0; i<nData; i++) {
      tR = cWork[2*i];
      cWork [2*i] = cWork[2*i+1];
      cWork [2*i+1] = tR;
   }
   *fft2 = minrtsMakeHermit (cWork, nData);
   free(work);
   free(cWork);
   return;
}  /* end of double real fft */

void minrtsDoubleRealFftB (int nData, float *fft1, float *fft2, 
   float **real1, float **real2) {
/* 
      do two hermitian backward ffts of the same length at the same time by putting one
      of them into the imaginary part of a complex number and use their
      hermitian properties to fish them out of the result 
*/
/* 
      note that the forward output has zero point in the middle, NOT AT EDGE 
*/
   int i, j;
   float *compX, *work, tR, tI;
   float *out1, *out2;
   compX = minrtsMakeHermit(fft1, nData);
   work = minrtsMakeHermit(fft2, nData);
/* 
      combine two complexes, shift zero point 
*/
   for (i=0; i<nData; i++) {
      compX[2*i]   =compX[2*i] - work[2*+1]; 
      compX[2*i+1] =compX[2*i+1] +  work[2*i];
   }
   for (i=0; i<nData/2; i++) {
      j = i + nData/2;
      tR = compX[2*i];
      tI = compX[2*i+1];
      compX[2*i]   = compX[2*j];
      compX[2*i+1] = compX[2*j+1];
      compX[2*j]   = tR;
      compX[2*j+1] = tI;
   }
   fourg (compX, nData, -1, work); /* fft */
   out1 = (float *)malloc(nData * sizeof(float));
   out2 = (float *)malloc(nData * sizeof(float));
   for (i=0; i<nData; i++) {
      out1[i] = compX[2*i]/nData;
      out2[i] = compX[2*i+1]/nData;
   }
   *real1 = out1;
   *real2 = out2;
   free(compX);
   free(work);
}  /* end of double real fft */

float *minrtsFftXShift(minrtsFft1d *ff, float *kData, float dx) {
/*  *assuming kData is the forward fft of something useful, shift its fft
    *by a phase ramp and transform it back                  
    * INPUTS:
    * ff         &    minrstfft object describing ffts
    * kData      &    complex fft data
    * dx         &    size of shift in units contained in ff
    * RETURNS:
    * output     &    float shifted fft of kData
    ************************************************************** 
   int i,j;
   int nX, nX2;
   float *compX, *output, tR, tI;
   nX  = ff->nX;
   nX2 = nX/2;
   output = (float *)malloc(nX *sizeof(float));
   for (i=0; i< nX; i++){
      j = i + nX2;
      if (j >= nX) j = j - nX;
      tR = cos(ff->kGrid[i] * dx);
      tI = sin(ff->kGrid[i] * dx);
      ff->cWork[2*j]   = tR * kData[2*i] - tI * kData[2*i+1];
      ff->cWork[2*j+1] = tI * kData[2*i] + tR * kData[2*i+1];
   }
   fourg (ff->cWork, nX, -1, ff->work);
   for (i=0; i< nX; i++){
      j = i + nX2;
      if (j >= nX) j = j - nX;
      output[i] = ff->cWork[2*j];
   }
   return output; */
  return NULL;
} /* end of xShift */

float minrtsDot(int nData, float* v1, float *v2) {
/* dot product of two vectors */
   int i;
   float output;
   output = 0;
   for (i=0;i<nData;i++) output = output + v1[i]*v2[i];
   return output;
} /* end minrtsDot */

void minrtsVCopy (int nData, float* vout, float *vin) {
/* vector copy */
   int i;
   for (i=0;i<nData;i++) vout[i] = vin[i];
   return;
} /* end of Vcopy */

void minrtsVAdd (int nData, float* v1, float *v2) {
/* vector add in place */
   int i;
   for (i=0;i<nData;i++) v1[i] = v1[i] + v2[i];
   return;
} /* end of VAdd */

void minrtsVMult (int nData, float* v1, float *v2) {
/* vector mult in place */
   int i;
   for (i=0;i<nData;i++) v1[i] = v1[i] * v2[i];
   return;
} /* end of VMult */

void minrtsVSMult (int nData, float* v1, float s) {
/* vector scalar mult in place */
   int i;
   for (i=0;i<nData;i++) v1[i] = v1[i] * s;
   return;
} /* end of VSMult */

void minrtsVSMultAdd (int nData, float* v1, float s1, float s2) {
/* vector scalar mult add in place */
   int i;
   for (i=0;i<nData;i++) v1[i] = v1[i] * s1 + s2;
   return;
} /* end of VSMultAdd */

float minrtsQPk (int nData, float *x, float *y, float *ymax, int domin) {
/* make quadratic fit to peak of y(x); return position of peak
   and value of peak */
   int i, j;
   float maxval, jmax, xm;
   maxval = y[0];
   j   = 0;
   /* find maximum or maximum, as requrested */
   if (domin == 0) {
      for (i=0; i<nData; i++) {
         if (y[i] > maxval) {
            maxval = y[i];
            j = i;
         }
      }
   }  /* integral pixel position of peak */
   else {
      for (i=0; i<nData; i++) {
         if (y[i] < maxval) {
            maxval = y[i];
            j   = i;  
         }
      } 
   }  /* integral pixel position of peak */
   j = vmax(j, 2);
   j = vmin(j, nData-3);
   jmax = -0.5*(y[j+1]-y[j-1])/(y[j+1]+y[j-1]-2*y[j]); /* fractional pixel position */
   xm = x[j] + .5*(x[j+1]-x[j-1])*jmax +
      .5*(x[j+1]+x[j-1]-2*x[j])*jmax*jmax; /* quad interpolated position of pk */
   *ymax = y[j] + .5*(y[j+1]-y[j-1])*jmax +
      .5*(y[j+1]+y[j-1]-2*y[j])*jmax*jmax; /* quad interpolated value of pk */
   return xm;
} /* end of minrtsQPk */

void minrtsExpC (int nData, float *phase, float *output) {

/* ******************************************************************
 * compute complex exponential of phase (i.e. exp(i*phase)) and
 * put answer in existing array output (float[2*nData])
 ******************************************************************* */
    int iData;
    for (iData =0; iData<nData; iData++) {
       output[2*iData]   = cos(phase[iData]);
       output[2*iData+1] = sin(phase[iData]);
    }
    return;
} /* end of minrtsExpC */


void minrtsCMult (int nData, float *comp1, float *comp2) {

/* ******************************************************************
 * Multiply (in place) two complex arrays, result overwrites 1st array
 ******************************************************************* */
    int iData;
    float x;
    for (iData =0; iData<nData; iData++) {
       x = comp1[2*iData]*comp2[2*iData] - comp1[2*iData+1]*comp2[2*iData+1];
       comp1[2*iData+1] = comp1[2*iData]*comp2[2*iData+1] + comp1[2*iData+1]*comp2[2*iData]; 
       comp1[2*iData] = x;
    }
    return;
} /* end of minrtsCMult */

void minrtsRCMult (int nData, float *real, float *comp) {

/* ******************************************************************
 * Multiply (in place) a real array times a complex array; 
 *   overwrites complex array
 ******************************************************************* */
    int iData;
    for (iData =0; iData<nData; iData++) {
       comp[2*iData]   = comp[2*iData] * real[iData];
       comp[2*iData+1] = comp[2*iData+1] * real[iData];
    }
    return;
} /* end of minrtsExpC */

void minrtsCReverse (int nData, float *comp) {

/* ******************************************************************
 * Reverse end to end (in place) a complex vector
 ******************************************************************* */
    int i;
    float x;
    for (i=0; i<nData/2; i++) {
       x = comp[2*i];
       comp[2*i] = comp[2*(nData-i-1)];
       comp[2*(nData-i-1)] = x;
       x = comp[2*i+1];
       comp[2*i+1] = comp[2*(nData-i)+1];
       comp[2*(nData-i)+1] = x;
    }
    return;
} /* end of minrtsCReverse */

void minrtsCConj (int nData, float *comp) {

/* ******************************************************************
 * complex conjugate
 ******************************************************************* */
    int i;
    for (i =0; i<nData; i++) 
       comp[2*i+1] = -comp[2*i+1];
    return;
} /* end of minrtsCConj */

void minrtsLinearFit(int nPoint, float *data, float *weight, float *fit){
   int i;
   float dC, dL, lL, lC, cC, x, y;
   float denom, numC, numL;
   dC = 0;
   dL = 0;
   lL = 0;
   lC = 0;
   cC = 0;
   for (i=0; i<nPoint;i++) {
      x = weight[i];
      y = (float)i;
      dC +=  x * data[i];
      dL +=  x * data[i] * y;
      lL +=  x * y * y;
      lC +=  x * y;
   }
   denom = lL*cC - lC*lC;
   numC  = (dC*lL - dL*lC)/denom;
   numL  = (cC*dL - lC*dC)/denom;
   for (i=0; i<nPoint;i++) fit[i] = numC + (float)i * numL;
   return;
} /* end of linear fit */


/*
 *        Fft based gaussian convolution of an array
 *        note that the smooth effect wraps around from end of array to beginning
 *        Inputs:
 *        nData   ; size of array
 *        data    ; input array
 *        output  ; preallocated output array
 *        sigma   ; std deviation in points of gaussian used for smooth
 *        ff      ; pointer to pointer to fft1d object.  Set contents to NULL 
 *                  for first call to routine, object will be created internally
 *                  use it again in subsequent calls;
 *        kill    ; if true destroy ff object before returning
*/               
void minrtsGaussSmooth(float sigma, int nData, float *data, float *output, 
   minrtsFft1d **ff, int killFft) { 

   int i, n2;
   float x, sig;
/*                 fft arrays if needed   */
   float gD[3];
   float *xPos;
/*           create fft object if not provided */
   if (ff[0] == NULL) {
      gD[0] = 0.;
      gD[1] = 1.;
      gD[2] = nData;
/*           create fft object */
      ff[0] = minrtsMakeFft1d(gD, 0) ;
      n2    = nData/2;
/*                  check sigma; set filter */
      if (sigma < .001) sig = .001; else sig = sigma;
      for (i=0;i<nData;i++)  {
         x = ff[0]->kGrid[i] * sig;
         x = x*x;
         if (x > 30.) x = 30.;
         x = exp(-0.5*x);
         ff[0]->filter[2*i  ] = x;
         ff[0]->filter[2*i+1] = 0.;
      }
   }
   xPos = (float *)malloc(nData * sizeof(float));
   for (i=0;i<nData;i++) xPos[i] = (float)i;

   minrtsFftFilterData (ff[0], nData, xPos, data, NULL, 0, 1, output) ;
   free(xPos);
   x = 0;
   if (killFft) {minrtsKillFft1d (ff[0]); ff[0] = NULL;}
   return;
}
